#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <string.h>

void initialize( float* x, int n )
{
  int n2 = n+2;
  for (int i=0; i<n2; ++i)
    for (int j=0; j<n2; ++j)
      x[i*n2 + j] = random() / (float) RAND_MAX;
}

__global__ void smooth( float* y, float* x, int n, float a, float b, float c )
{
    int n2 = n+2, i, j, k, l, block=18;
   
     __shared__ float xb[18*18];

    k = blockDim.x * blockIdx.x + threadIdx.x + 1; // add one to start from index 1
    l = blockDim.y * blockIdx.y + threadIdx.y + 1;
    i = threadIdx.x + 1;
    j = threadIdx.y + 1;

    /* copy data into shared buffer */
    if (i == 1 & j==1)
    {
        for (int ii=0; ii <= (blockDim.x+1); ++ii)
            for (int jj=0; jj <= (blockDim.y+1); ++jj)
                xb[ii*block + jj] = x[(k+ii-1)*n2 + (l+jj-1)];
    }
    __syncthreads();

    y[l*n2 + k] = a * (xb[(j-1)*block+(i-1)] + xb[(j-1)*block+(i+1)] + 
                       xb[(j+1)*block+(i-1)] + xb[(j+1)*block + (i+1)])\
                + b * (xb[j*block + (i-1)] + xb[j*block + (i+1)] + 
                       xb[(j-1)*block + i] + xb[(j+1)*block + i])
                + c * xb[j*block + i];
}

__global__ void smooth16( float* y, float* x, int n, float a, float b, float c )
{
    int n2 = n+2, i, j, k, l, block=18;
    
     __shared__ float xb[18*18];
  
    k = blockDim.x * blockIdx.x + threadIdx.x + 1; // add one to start from index 1
    l = blockDim.y * blockIdx.y + threadIdx.y + 1;
    i = threadIdx.x + 1;
    j = threadIdx.y + 1;
        
    /* copy inner part: block x block elements */
    xb[j*block + i] = x[l*n2 + k];
  
    /* copy boundaries */
    /* x and corners */
    if (i == 1)
      {
      xb[j*block + (i-1)] = x[l*n2 + (k-1)];
      if (j == 1)
        {
        xb[(j-1)*block + (i-1)] = x[(l-1)*n2 + (k-1)];
        xb[(j+1)*block + (i-1)] = x[(l+1)*n2 + (k-1)];
        }
      else if (j == blockDim.y)
       {
       xb[(j+1)*block + (i-1)] = x[(l+1)*n2 + (k-1)];
       }
     }
   else if (i == blockDim.x)
     {
     xb[j*block + (i+1)] = x[l*n2 + (k+1)];
     if (j == 1)
       {
       xb[(j-1)*block + (i+1)] = x[(l-1)*n2 + (k+1)];
       xb[(j+1)*block + (i+1)] = x[(l+1)*n2 + (k+1)];
       }
     else if (j == blockDim.y)
       {
       xb[(j+1)*block + (i+1)] = x[(l+1)*n2 + (k+1)];
       }
     }
  /* y */
  if (j == 1)
    xb[(j-1)*block + i] = x[(l-1)*n2 + k];
  else if (j == blockDim.y)
    xb[(j+1)*block + i] = x[(l+1)*n2 + k];

    __syncthreads();
    
    y[l*n2 + k] = a * (xb[(j-1)*block+(i-1)] + xb[(j-1)*block+(i+1)] +
                       xb[(j+1)*block+(i-1)] + xb[(j+1)*block + (i+1)])\
                + b * (xb[j*block + (i-1)] + xb[j*block + (i+1)] +
                       xb[(j-1)*block + i] + xb[(j+1)*block + i])
                + c * xb[j*block + i];
}

__global__ void smooth18( float* y, float* x, int n, float a, float b, float c )
{
    int n2 = n+2, i, j, k, l, lind;
    
     __shared__ float xb[18*18];
  
    k = 16 * blockIdx.x + threadIdx.x; // add one to start from index 1
    l = 16 * blockIdx.y + threadIdx.y;
    i = threadIdx.x + 1;
    j = threadIdx.y + 1;
    int block = 18; 

    /* copy tile to shared memory */
    xb[threadIdx.y*18 + threadIdx.x] = x[l*n2 + k];

    __syncthreads();

    /* compute linear thread index */
    lind = blockDim.x * threadIdx.y + threadIdx.x + 1;

    if (lind <= 256)
    {
        i = (lind-1) / 16 + 1;
        j = lind - (i-1) * 16;
        k = 16 * blockIdx.x + i;
        l = 16 * blockIdx.y + j;

    
    y[l*n2 + k] = a * (xb[(j-1)*block+(i-1)] + xb[(j-1)*block+(i+1)] +
                       xb[(j+1)*block+(i-1)] + xb[(j+1)*block + (i+1)])\
                + b * (xb[j*block + (i-1)] + xb[j*block + (i+1)] +
                       xb[(j-1)*block + i] + xb[(j+1)*block + i])
                + c * xb[j*block + i];
    }
}
void count( float* x, int n, float t, int &nbx )
{
  nbx = 0;
  int n2 = n+2;
  for (int i=1; i <= n; ++i)
    for (int j=1; j <= n; ++j)
      if (x[i*n2 + j] < t)
        ++nbx;
}

class Timer
{
public:
  Timer(): n(0) { }
  void start(std::string label) 
  { 
    if (n < 20) 
      { labels[n] = label; times[2*n] = clock(); }
    else { std::cerr << "No more timers, " << label 
      << " will not be timed." << std::endl; }
  }

  void stop() { times[2*n+1] = clock(); n++;}
  void reset() { n=0; }
  void print();
private:
  std::string labels[20];
  float times[40];
  int n;
};


void Timer::print()
{
    std::cout << std::endl;
    std::cout << "Action           ::   time/s     Time resolution = " << 1.f/(float)CLOCKS_PER_SEC << std::endl;
    std::cout << "------" << std::endl;
    for (int i=0; i < n; ++i)
        std::cout << labels[i] << " :: " << (times[2*i+1] - times[2*i+0])/(float)CLOCKS_PER_SEC << std::endl;
}

int main()
{

int n,nbx,nby;

float a,b,c,t,*x,*y;

/* timer of class Timer */
Timer timer, timer2;

/* n - number of elements in one direction */
    n = 1 << 14;  /* 2^14 == 16384 */

/* a,b,c - smoothing constants */
    a = 0.05;
    b = 0.1;
    c = 0.4;

/* t - threshold */
    t = 0.5;

/* allocate x */
    x = new float[(n+2)*(n+2)];

/* allocate y */
    y = new float[(n+2)*(n+2)];

 /* initialize x */
initialize(x, n);
    
/* smooth x into y */
//this can be changed to smooth, smooth16, smooth18 to change between one thread copying all, 16 threads with duplicates, 18 threads with one each, respectively
/* launch smooth() on GPU */
//int nBlocks = 16;
int nBlocks = 18;
timer.start("GPU: Smooth     ");
//smooth<<<n/nBlocks,nBlocks>>>(y, x, n, a, b, c);
//smooth16<<<n/nBlocks,nBlocks>>>(y, x, n, a, b, c);
smooth18<<<n/nBlocks,nBlocks>>>(y, x, n, a, b, c);
cudaDeviceSynchronize();
timer.stop();

/* count elements in second array */
timer2.start("CPU: Count-XY");
  timer.start("CPU: Count-X");
  count(x, n, t, nbx);
  timer.stop(); 
  timer.start("CPU: Count-Y");
  count(y, n, t, nby);
  timer.stop();
timer2.stop();

std::cout << std::endl;
std::cout << "Summary" << std::endl;
std::cout << "-------" << std::endl;
std::cout << "Number of elements in a row/column         :: " << n+2 << std::endl;
std::cout << "Number of inner elements in a row/column   :: " << n << std::endl;
std::cout << "Total number of elements                   :: " << (n+2)*(n+2) << std::endl;
std::cout << "Total number of inner elements             :: " << n*n << std::endl;
std::cout << "Memory (GB) used per array                 :: " << (n+2)*(n+2)*sizeof(float) / (float)(1024*1024*1024) << std::endl;;
std::cout << "Threshold                                  :: " << t << std::endl;
std::cout << "Smoothing constants (a, b, c)              :: " << a << " " << b << " " << c << std::endl;
std::cout << "Number   of elements below threshold (X)   :: " << nbx << std::endl;
std::cout << "Fraction of elements below threshold       :: " << nbx / (float)(n*n) << std::endl;
std::cout << "Number   of elements below threshold (Y)   :: " << nby << std::endl;
std::cout << "Fraction of elements below threshold       :: " << nby / (float)(n*n) << std::endl;

timer.print();
timer2.print();

return(0);
}
